@@ -719,3 +719,13 @@ def big_number_pretty_print(n):
     ret.reverse()
 
     return ''.join(ret[1:]) if ret[0] == ',' else ''.join(ret)
+
+def combine_files(in_paths, out_path):
+    with open(out_path, 'w') as out_file:
+        for in_path in in_paths:
+            try:
+                with open(in_path) as in_file:
+                    for line in in_file:
+                        out_file.write(line)
+            except FileNotFoundError:
+                pass
\ No newline at end of file
